/* classe che realizza la gestione degli accessi alla blackboard e la gestione
del passaggio da Java al motore Prolog dell'agente e viceversa */

import amzi.ls.*;
import java.util.Vector;
import java.rmi.*;
import java.rmi.registry.*;
import java.rmi.server.*;
import java.io.*;

class logicAgent
{
   LogicServer ls; /* motore Prolog */
   int ist;
   AgentFrame referenceFrame=null; /* riferimento all'oggetto interfaccia dell'agente */
   agentServer server=null;  /* riferimento all'oggetto server dell'agente */


   public logicAgent(AgentFrame referenceFrame,agentServer server,int id) throws LSException
   {
      ist=id;
      this.referenceFrame=referenceFrame;
      this.server=server;
      try {
        // create and initialize a logic server
        ls = new LogicServer();
        ls.Init("");

        // add extended predicates
        ls.AddPred("ask_remote_cons",3,"logicAgent","askConsistency",this);
        ls.AddPred("prompt",1,"logicAgent","prompt",this);
        ls.AddPred("aggiorna",1,"logicAgent","aggiorna",this);

        // load the compiled Prolog expert system and the metainterpreter
        ls.Load(server.rules);
      } catch (LSException e) {System.out.println(e.GetMsg());}
      System.out.println("Istanziato il logic Server e caricato il database");
   }

/* metodo che gestisce la derivazione abduttiva */
   public String solveAbducible(String goal) throws LSException
   {
      String ris=null;
      try {
        /* Caricamento metainterprete */
        long s=ls.ExecStr("load([startabducible])");
        long t,term,risultato;
        String abducibili;
        /* Accesso alla blackboard */
        if (server.politicaInOut) {
           abducibili=server.blackboard.In();
        } else {
           abducibili=server.blackboard.read();
        }
        /* Passaggio dell'esecuzione al motore Prolog: invocazione query */
        term=ls.ExecStr("start_abductive_derivation(["+goal+"],"+abducibili+",AbdOut)");
        /* Termine esecuzione Prolog: ritornata una risposta
         true --> term>0, false --> term=0 */
        if (term != 0) {
           risultato=ls.GetArg(term,3);   //risultato contiene la lista degli abducibili trovati
           ris=ls.TermToStr(risultato,10000);
           /* Reinserimento valore nella blackboard in caso di gestione con la politica In/Out;
           in caso di politica read/write e' effettuata direttamente dal metainterprete Prolog */
           if (server.politicaInOut) {
              server.blackboard.Out(ris);
           }
        } else if (server.politicaInOut) {
                  server.blackboard.Out(abducibili);
               }
      } catch (Exception e) {}
      return ris;
   }

/* metodo che gestice la derivazione di consistenza */
   public Vector askConsLogicAgent(String goal,Vector abdListIn) throws LSException
   {
      try {
      /* Caricamento metainterprete */
      long t=ls.ExecStr("load([startabducible])");
      } catch (LSException e) {System.out.println(e.GetMsg());}
      long abdin,term,risultato,result;
      Vector v=new Vector();
      String ris=null;
      String abdStr=null;
      /* Preparazione argomenti della richiesta di consistenza */
      abdin=ls.MakeList();
      for(int i=0;i<abdListIn.size();i++) {
        term=ls.StrToTerm((String)abdListIn.elementAt(i));
        abdin=ls.PushList(abdin,term);
      }           //abdin contiene la lista degli abducibili in ingresso
      abdStr=ls.TermToStr(abdin,10000);
      /* Passaggio dell'esecuzione al motore Prolog: invocazione query */
      term = ls.ExecStr("solve_con("+goal+","+abdStr+",AbdOut)");
      /* Termine esecuzione Prolog: ritornata una risposta
         true --> term>0, false --> term=0 */
      if (term > 0)
        {
          risultato=ls.GetArg(term, 3);  //risultato contiene la lista degli abducibili trovati
          String risultatoStr=ls.TermToStr(risultato,10000);
          while (risultato!=0) {
            result=ls.GetHead(risultato);
            ris=ls.TermToStr(result,100);
            v.addElement(ris);
            risultato=ls.GetTail(risultato);
          }
          // v rappresenta il vettore degli abducibili trovati
        }
      else
        {
         //La query e' fallita: il vettore di ritorno v e' vuoto
        }

      return v;
   }

 /* metodo richiamato dal motore Prolog quando viene incontrato il predicato
 ask_remote_cons(Goal,AbdIn,AbdOut) che fa tornare il controllo a Java */
   public boolean askConsistency() throws LSException
   {
      Vector abducibili=new Vector();
      Vector abdOut=new Vector();
      String abd=null;
      long head;
      /* Conversione argomenti Prolog in tipi Java */
      long term=ls.GetParm(1);
      String goal=ls.TermToStr(term,100);   //goal e' l'abducibile su cui verificare la consistency
      System.out.println("Gol da ask_consistency="+goal);
      long lista=ls.GetParm(2); //lista contiene la lista degli abducibili trovati fin'ora
      String listastr=ls.TermToStr(lista,10000);
      System.out.println("Lista abdIn vale "+listastr);
      while (lista!=0) {
         head=ls.GetHead(lista);
         abd=ls.TermToStr(head,100);
         abducibili.addElement(abd);
         lista=ls.GetTail(lista);
      }
      iAgent remoteCall;
      clientThread t=null;
      Vector waitThread=new Vector(server.numberOfAgent);
      Vector risposta=new Vector();
      String serv=null;
      try {
        /* Avvio di un thread per ogni agente del gruppo */
        for (int i=0;i<server.numberOfAgent;i++) {
          serv="rmi://"+server.otherMachine.elementAt(i)+":"+server.otherRegistry.elementAt(i)+"/"+server.otherAgentName.elementAt(i);
          remoteCall=(iAgent)Naming.lookup(serv);
          referenceFrame.printMessage("Invio richiesta di consistenza per il termine '"+goal+"' all'agente "+serv);
          String name="'"+(i+1)+"-a chiamata remota a' "+server.otherAgentName.elementAt(i);
          t=new clientThread(name,remoteCall,goal,abducibili);
          t.start();
          waitThread.addElement(t);
        }
        try {
          /* Gestione delle risposte ottenute da ciascun thread */
          for (int i=0;i<server.numberOfAgent;i++) {
            clientThread w=(clientThread)waitThread.firstElement();
            waitThread.removeElementAt(0);
            w.join();
            risposta=w.getResult();
            serv="rmi://"+server.otherMachine.elementAt(i)+":"+server.otherRegistry.elementAt(i)+"/"+server.otherAgentName.elementAt(i);
            if (risposta.size()>0) {
                referenceFrame.printVectorMessage("Ricevuta risposta per la consistenza del termine '"
                                          +goal+"' dall'agente "+serv+"\nOttenuta la lista risultato ",risposta);
                for (int j=0;j<risposta.size();j++) {
                    abdOut.addElement(risposta.elementAt(j));
                }
            } else {
                referenceFrame.printMessage("La richiesta di consistenza per il termine '"+goal+"' all'agente "
                             +serv+" e' fallita");
                for (int j=i+1;j<server.numberOfAgent;j++) {
                    w=(clientThread)waitThread.firstElement();
                    waitThread.removeElementAt(0);
                    if (w.isAlive()) {
                        System.out.println("interrotto il thread "+w.getName());
                        w.stop();
                    } else { System.out.println("il thread "+w.getName()+" e' gia' terminato"); }
                }
                return false;
            }
          }
        } catch (Exception ignored) {System.out.println("join fallita");}
      } catch (Exception e) {e.printStackTrace();}
      long termList=ls.MakeList();
      long termine;
      for(int i=0;i<abdOut.size();i++) {
        termine=ls.StrToTerm((String)abdOut.elementAt(i));
        termList=ls.PushList(termList,termine);
      }
      ls.UnifyParm(3,termList);
      return true;
   }

/* metodo richiamato dal motore Prolog quando viene incontrato il predicato prompt/1 */
   public synchronized boolean prompt() throws LSException
   {
       long t=ls.GetParm(1);
       String s=ls.TermToStr(t,10000);
       referenceFrame.textAreaMessages.append("PROLOG (istanza "+ist+"): "+s+"\n");
       return true;
   }

/* metodo richiamato dal motore Prolog quando viene incontrato il predicato aggiorna/1
per l'aggiornamento della blackboard nel caso di gestione tramite la politica read/write */
   public synchronized boolean aggiorna() throws LSException
   {
       boolean res=false;
       long lista=ls.GetParm(1);  //lista contiene la lista degli abducibili trovati
       String listastr=ls.TermToStr(lista,10000);
       System.out.println("Chiamato il predicato aggirna: cosa succede?");
       try {
         if (server.politicaInOut) {
            res=true;
         } else {
             res=server.blackboard.write(listastr);
             if (!res) {
                referenceFrame.printMessage("Il risultato trovato e' in conflitto con quello presente nella blackboard");
                System.out.println("Si prova a cercarne uno nuovo");
             } else { System.out.println("La write dovrebbe essere andata bene");}
         }
       } catch (RemoteException e) {}
       return res;
   }

   public void close() throws LSException
   {
      if (ls != null)
         ls.Close();
   }

}